"""Tests for Pydantic utility functions."""

from __future__ import annotations

import json
from datetime import UTC, datetime, timedelta, timezone

import pytest
from pydantic import BaseModel, ValidationError, root_validator

from safir.pydantic import (
    CamelCaseModel,
    normalize_datetime,
    normalize_isodatetime,
    to_camel_case,
    validate_exactly_one_of,
)


def test_normalize_datetime() -> None:
    assert normalize_datetime(None) is None

    date = datetime.fromtimestamp(1668814932, tz=UTC)
    assert normalize_datetime(1668814932) == date

    mst_zone = timezone(-timedelta(hours=7))
    mst_date = datetime.now(tz=mst_zone)
    utc_date = mst_date.astimezone(UTC)
    assert normalize_datetime(mst_date) == utc_date

    naive_date = datetime.utcnow()  # noqa: DTZ003
    aware_date = normalize_datetime(naive_date)
    assert aware_date == naive_date.replace(tzinfo=UTC)
    assert aware_date.tzinfo == UTC


def test_normalize_isodatetime() -> None:
    assert normalize_isodatetime(None) is None

    date = datetime.fromisoformat("2023-01-25T15:44:34+00:00")
    assert date == normalize_isodatetime("2023-01-25T15:44:34Z")

    date = datetime.fromisoformat("2023-01-25T15:44:00+00:00")
    assert date == normalize_isodatetime("2023-01-25T15:44Z")

    with pytest.raises(ValueError, match=r"Must be a string in .* format"):
        normalize_isodatetime("2023-01-25T15:44:00+00:00")

    with pytest.raises(ValueError, match=r"Must be a string in .* format"):
        normalize_isodatetime(1668814932)  # type: ignore[arg-type]

    with pytest.raises(ValueError, match=r"Must be a string in .* format"):
        normalize_isodatetime("next thursday")


def test_to_camel_case() -> None:
    assert to_camel_case("foo") == "foo"
    assert to_camel_case("minimum_lifetime") == "minimumLifetime"
    assert to_camel_case("replace_403") == "replace403"
    assert to_camel_case("foo_bar_baz") == "fooBarBaz"


def test_camel_case_model() -> None:
    class TestModel(CamelCaseModel):
        minimum_lifetime: int
        replace_403: bool
        foo_bar_baz: str

    camel = {
        "minimumLifetime": 10,
        "replace403": False,
        "fooBarBaz": "something",
    }
    snake = {
        "minimum_lifetime": 10,
        "replace_403": False,
        "foo_bar_baz": "something",
    }
    data = TestModel.parse_obj(camel)
    assert data.minimum_lifetime == 10
    assert not data.replace_403
    assert data.foo_bar_baz == "something"
    assert data.dict() == camel
    assert data.dict(by_alias=False) == snake
    assert data.json() == json.dumps(camel)
    assert data.json(by_alias=False) == json.dumps(snake)

    snake_data = TestModel.parse_obj(snake)
    assert data.minimum_lifetime == 10
    assert not data.replace_403
    assert data.foo_bar_baz == "something"
    assert snake_data.dict() == data.dict()
    assert snake_data.json() == data.json()


def test_validate_exactly_one_of() -> None:
    class Model(BaseModel):
        foo: int | None = None
        bar: int | None = None
        baz: int | None = None

        _validate_type = root_validator(allow_reuse=True)(
            validate_exactly_one_of("foo", "bar", "baz")
        )

    Model.parse_obj({"foo": 4, "bar": None})
    Model.parse_obj({"baz": 4})
    Model.parse_obj({"bar": 4})
    Model.parse_obj({"foo": None, "bar": 4})

    with pytest.raises(ValidationError) as excinfo:
        Model.parse_obj({"foo": 4, "bar": 3, "baz": None})
    assert "only one of foo, bar, and baz may be given" in str(excinfo.value)

    with pytest.raises(ValidationError) as excinfo:
        Model.parse_obj({"foo": None, "baz": None})
    assert "one of foo, bar, and baz must be given" in str(excinfo.value)

    class TwoModel(BaseModel):
        foo: int | None = None
        bar: int | None = None

        _validate_type = root_validator(allow_reuse=True)(
            validate_exactly_one_of("foo", "bar")
        )

    with pytest.raises(ValidationError) as excinfo:
        TwoModel.parse_obj({"foo": 3, "bar": 4})
    assert "only one of foo and bar may be given" in str(excinfo.value)

    with pytest.raises(ValidationError) as excinfo:
        TwoModel.parse_obj({})
    assert "one of foo and bar must be given" in str(excinfo.value)



"""Date and time manipulation utility functions."""

from __future__ import annotations

from datetime import UTC, datetime
from typing import overload

__all__ = [
    "current_datetime",
    "format_datetime_for_logging",
    "isodatetime",
    "parse_isodatetime",
]


def current_datetime(*, microseconds: bool = False) -> datetime:
    """Construct a `~datetime.datetime` for the current time.

    It's easy to forget to force all `~datetime.datetime` objects to be time
    zone aware. This function forces UTC for all objects.

    Databases do not always store microseconds in time fields, and having some
    dates with microseconds and others without them can lead to bugs, so by
    default it suppresses the microseconds.

    Parameters
    ----------
    microseconds
        Whether to include microseconds. Consider setting this to `True` when
        getting timestamps for error reporting, since granular timestamps can
        help in understanding sequencing.

    Returns
    -------
    datetime.datetime
        The current time forced to UTC and optionally with the microseconds
        field zeroed.
    """
    result = datetime.now(tz=UTC)
    if microseconds:
        return result
    else:
        return result.replace(microsecond=0)


@overload
def format_datetime_for_logging(timestamp: datetime) -> str:
    ...


@overload
def format_datetime_for_logging(timestamp: None) -> None:
    ...


def format_datetime_for_logging(timestamp: datetime | None) -> str | None:
    """Format a datetime for logging and human readabilty.

    Parameters
    ----------
    timestamp
        Object to format. Must be in UTC or timezone-naive (in which case it's
        assumed to be in UTC).

    Returns
    -------
    str or None
        The datetime in format ``YYYY-MM-DD HH:MM:SS[.sss]`` with milliseconds
        added if and only if the microseconds portion of ``timestamp`` is not
        0. There will be no ``T`` separator or time zone information.

    Raises
    ------
    ValueError
        Raised if the argument is in a time zone other than UTC.
    """
    if timestamp:
        if timestamp.tzinfo not in (None, UTC):
            raise ValueError(f"datetime {timestamp} not in UTC")
        if timestamp.microsecond:
            result = timestamp.isoformat(sep=" ", timespec="milliseconds")
        else:
            result = timestamp.isoformat(sep=" ", timespec="seconds")
        return result.split("+")[0]
    else:
        return None


def isodatetime(timestamp: datetime) -> str:
    """Format a timestamp in UTC in a standard ISO date format.

    Parameters
    ----------
    timestamp
        Date and time to format.

    Returns
    -------
    str
        Date and time formatted as an ISO 8601 date and time using ``Z`` as
        the time zone.  This format is compatible with both Kubernetes and the
        IVOA UWS standard.

    Raises
    ------
    ValueError
        The provided timestamp was not in UTC.
    """
    if timestamp.tzinfo not in (None, UTC):
        raise ValueError(f"datetime {timestamp} not in UTC")
    return timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")


def parse_isodatetime(time_string: str) -> datetime | None:
    """Parse a string in a standard ISO date format.

    Parameters
    ----------
    time_string
        Date and time formatted as an ISO 8601 date and time using ``Z`` as
        the time zone.  This is the same format produced by `isodatetime` and
        is compatible with Kubernetes and the IVOA UWS standard.

    Returns
    -------
    datetime.datetime
        The corresponding `datetime.datetime`.

    Raises
    ------
    ValueError
        The provided ``time_string`` is not in the correct format.

    Notes
    -----
    When parsing input for a model, use `safir.pydantic.normalize_isodatetime`
    instead of this function.  Using a model will be the normal case; this
    function is primarily useful in tests.
    """
    if not time_string.endswith("Z"):
        raise ValueError(f"{time_string} does not end with Z")
    return datetime.fromisoformat(time_string[:-1] + "+00:00")



"""Gafaelfawr authentication dependencies."""

from fastapi import Depends, Header
from structlog.stdlib import BoundLogger

from .logger import logger_dependency

__all__ = [
    "auth_dependency",
    "auth_logger_dependency",
]


async def auth_dependency(
    x_auth_request_user: str = Header(..., include_in_schema=False)
) -> str:
    """Retrieve authentication information from HTTP headers.

    Intended for use with applications protected by Gafaelfawr, this retrieves
    authentication information from headers added to the incoming request by
    the Gafaelfawr ``auth_request`` NGINX subhandler.
    """
    return x_auth_request_user


async def auth_logger_dependency(
    user: str = Depends(auth_dependency),
    logger: BoundLogger = Depends(logger_dependency),
) -> BoundLogger:
    """Logger bound to the authenticated user."""
    return logger.bind(user=user)



"""Middleware for IVOA services."""

from collections.abc import Awaitable, Callable
from urllib.parse import urlencode

from fastapi import Request, Response
from starlette.middleware.base import BaseHTTPMiddleware

__all__ = ["CaseInsensitiveQueryMiddleware"]


class CaseInsensitiveQueryMiddleware(BaseHTTPMiddleware):
    """Make query parameter keys all lowercase.

    Unfortunately, several IVOA standards require that query parameters be
    case-insensitive, which is not supported by modern HTTP web frameworks.
    This middleware attempts to work around this by lowercasing the query
    parameter keys before the request is processed, allowing normal FastAPI
    query parsing to then work without regard for case.  This, in turn,
    permits FastAPI to perform input validation on GET parameters, which would
    otherwise only happen if the case used in the request happened to match
    the case used in the function signature.

    This unfortunately doesn't handle POST, so routes that accept POST will
    need to parse the POST data case-insensitively in the handler or a
    dependency.

    Based on `fastapi#826 <https://github.com/tiangolo/fastapi/issues/826>`__.
    """

    async def dispatch(
        self,
        request: Request,
        call_next: Callable[[Request], Awaitable[Response]],
    ) -> Response:
        params = [(k.lower(), v) for k, v in request.query_params.items()]
        request.scope["query_string"] = urlencode(params).encode()
        return await call_next(request)



"""Utility functions for Pydantic models."""

from __future__ import annotations

from collections.abc import Callable
from datetime import UTC, datetime
from typing import Any, ParamSpec, TypeVar

from pydantic import BaseModel

P = ParamSpec("P")
T = TypeVar("T")

__all__ = [
    "CamelCaseModel",
    "normalize_datetime",
    "normalize_isodatetime",
    "to_camel_case",
    "validate_exactly_one_of",
]


def normalize_datetime(v: int | datetime | None) -> datetime | None:
    """Pydantic validator for datetime fields.

    Supports `~datetime.datetime` fields given in either any format supported
    by Pydantic natively, or in seconds since epoch (which Pydantic doesn't
    support).  This validator ensures that datetimes are always stored in the
    model as timezone-aware UTC datetimes.

    Parameters
    ----------
    v
        The field representing a `~datetime.datetime`.

    Returns
    -------
    datetime.datetime or None
        The timezone-aware `~datetime.datetime` or `None` if the input was
        `None`.

    Examples
    --------
    Here is a partial model that uses this function as a validator.

    .. code-block:: python

       class Info(BaseModel):
           last_used: Optional[datetime] = Field(
               None,
               title="Last used",
               description="When last used in seconds since epoch",
               example=1614986130,
           )

           _normalize_last_used = validator(
               "last_used", allow_reuse=True, pre=True
           )(normalize_datetime)
    """
    if v is None:
        return v
    elif isinstance(v, int):
        return datetime.fromtimestamp(v, tz=UTC)
    elif v.tzinfo and v.tzinfo.utcoffset(v) is not None:
        return v.astimezone(UTC)
    else:
        return v.replace(tzinfo=UTC)


def normalize_isodatetime(v: str | None) -> datetime | None:
    """Pydantic validator for datetime fields in ISO format.

    This validator requires the ISO 8601 date and time format with ``Z`` as
    the time zone (``YYYY-MM-DDTHH:MM:SSZ``).  This format is compatible with
    Kubernetes and the ISO UWS standard and is the same format produced by
    `safir.datetime.isodatetime`.  It should be used when the ambiguous
    formats supported by Pydantic by default (such as dates and times without
    time zone information) shouldn't be allowed.

    Parameters
    ----------
    v
        The field representing a `~datetime.datetime`.

    Returns
    -------
    datetime.datetime or None
        The timezone-aware `~datetime.datetime` or `None` if the input was
        `None`.

    Examples
    --------
    Here is a partial model that uses this function as a validator.

    .. code-block:: python

       class Info(BaseModel):
           last_used: Optional[datetime] = Field(
               None,
               title="Last used",
               description="Date and time last used",
               example="2023-01-25T15:44:34Z",
           )

           _normalize_last_used = validator(
               "last_used", allow_reuse=True, pre=True
           )(normalize_isodatetime)
    """
    if v is None:
        return None
    if not isinstance(v, str) or not v.endswith("Z"):
        raise ValueError("Must be a string in YYYY-MM-DDTHH:MM[:SS]Z format")
    try:
        return datetime.fromisoformat(v[:-1] + "+00:00")
    except Exception as e:
        raise ValueError(f"Invalid date {v}: {e!s}") from e


def to_camel_case(string: str) -> str:
    """Convert a string to camel case.

    Intended for use with Pydantic as an alias generator so that the model can
    be initialized from camel-case input, such as Kubernetes objects or
    settings from Helm charts.

    Parameters
    ----------
    string
        Input string.

    Returns
    -------
    str
        String converted to camel-case with the first character in lowercase.

    Examples
    --------
    To support ``camelCase`` input to a model, use the following settings:

    .. code-block:: python

       class Model(BaseModel):
           some_field: str

           class Config:
               alias_generator = to_camel_case
               allow_population_by_field_name = True

    This must be added to every class that uses ``snake_case`` for an
    attribute and that needs to be initialized from ``camelCase``.
    Alternately, inherit from `~safir.pydantic.CamelCaseModel`, which is
    derived from `pydantic.BaseModel` with those settings added.
    """
    components = string.split("_")
    return components[0] + "".join(c.title() for c in components[1:])


def _copy_type(
    parent: Callable[P, T]
) -> Callable[[Callable[..., T]], Callable[P, T]]:
    """Copy the type of a parent method.

    Used to avoid duplicating the prototype of `pydantic.BaseModel.dict` and
    `pydantic.BaseModel.json` when overriding them in `CamelCaseModel`.

    Parameters
    ----------
    parent
        Method from which to copy a type signature.

    Returns
    -------
    Callable
        Decorator that will replace the type signature of a function with the
        type signature of its parent.
    """
    return lambda x: x


class CamelCaseModel(BaseModel):
    """`pydantic.BaseModel` configured to accept camel-case input.

    This is a convenience class identical to `~pydantic.BaseModel` except with
    an alias generator configured so that it can be initialized with either
    camel-case or snake-case keys. Model exports with ``dict`` or ``json``
    also default to exporting in camel-case.
    """

    class Config:
        alias_generator = to_camel_case
        allow_population_by_field_name = True

    @_copy_type(BaseModel.dict)
    def dict(self, **kwargs: Any) -> dict[str, Any]:
        """Export the model as a dictionary.

        Overridden to change the default of ``by_alias`` from `False` to
        `True`, so that by default the exported dictionary uses camel-case.
        """
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        return super().dict(**kwargs)

    @_copy_type(BaseModel.json)
    def json(self, **kwargs: Any) -> str:
        """Export the model as JSON.

        Overridden to change the default of ``by_alias`` from `False` to
        `True`, so that by default the exported dictionary uses camel-case.
        """
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        return super().json(**kwargs)


def validate_exactly_one_of(
    *settings: str,
) -> Callable[[type, dict[str, Any]], dict[str, Any]]:
    """Generate a validator imposing a one and only one constraint.

    Sometimes, models have a set of attributes of which one and only one may
    be set. Ideally this is represented properly in the type system, but
    occasionally it's more convenient to use a validator. This is a validator
    generator that can produce a validator function that ensures one and only
    one of an arbitrary set of attributes must be set.

    Parameters
    ----------
    *settings
        List of names of attributes, of which one and only one must be set.
        At least two attribute names must be listed.

    Returns
    -------
    Callable
        The root validator.

    Examples
    --------
    Use this inside a Pydantic class as a validator as follows:

    .. code-block:: python

       class Foo(BaseModel):
           foo: Optional[str] = None
           bar: Optional[str] = None
           baz: Optional[str] = None

           _validate_options = root_validator(allow_reuse=True)(
               validate_exactly_one_of("foo", "bar", "baz")
           )

    The attribute listed as the first argument to the ``validator`` call must
    be the last attribute in the model definition so that any other attributes
    have already been seen.
    """
    if len(settings) < 2:
        msg = "validate_exactly_one_of takes at least two field names"
        raise ValueError(msg)

    if len(settings) == 2:
        options = f"{settings[0]} and {settings[1]}"
    else:
        options = ", ".join(settings[:-1]) + ", and " + settings[-1]

    def validator(cls: type, values: dict[str, Any]) -> dict[str, Any]:
        seen = False
        for setting in settings:
            if setting in values and values[setting] is not None:
                if seen:
                    raise ValueError(f"only one of {options} may be given")
                seen = True
        if not seen:
            raise ValueError(f"one of {options} must be given")
        return values

    return validator



"""Update the request based on ``X-Forwarded-For`` headers."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from ipaddress import _BaseAddress, _BaseNetwork, ip_address

from fastapi import FastAPI, Request, Response
from starlette.middleware.base import BaseHTTPMiddleware

__all__ = ["XForwardedMiddleware"]


class XForwardedMiddleware(BaseHTTPMiddleware):
    """Middleware to update the request based on ``X-Forwarded-For``.

    The remote IP address will be replaced with the right-most IP address in
    ``X-Forwarded-For`` that is not contained within one of the trusted
    proxy networks.

    If ``X-Forwarded-For`` is found and ``X-Forwarded-Proto`` is also present,
    the corresponding entry of ``X-Forwarded-Proto`` is used to replace the
    scheme in the request scope.  If ``X-Forwarded-Proto`` only has one entry
    (ingress-nginx has this behavior), that one entry will become the new
    scheme in the request scope.

    The contents of ``X-Forwarded-Host`` will be stored as ``forwarded_host``
    in the request state if it and ``X-Forwarded-For`` are present.  Normally
    this is not needed since NGINX will pass the original ``Host`` header
    without modification.

    Parameters
    ----------
    proxies
        The networks of the trusted proxies.  If not specified, defaults to
        the empty list, which means only the immediately upstream proxy will
        be trusted.
    """

    def __init__(
        self, app: FastAPI, *, proxies: list[_BaseNetwork] | None = None
    ) -> None:
        super().__init__(app)
        if proxies:
            self.proxies = proxies
        else:
            self.proxies = []

    async def dispatch(
        self,
        request: Request,
        call_next: Callable[[Request], Awaitable[Response]],
    ) -> Response:
        """Middleware to update the request based on ``X-Forwarded-For``.

        Parameters
        ----------
        request
            The incoming request.
        call_next
            The next step in the processing stack.

        Returns
        -------
        ``fastapi.Response``
            The response with additional information about proxy headers.
        """
        forwarded_for = list(reversed(self._get_forwarded_for(request)))
        if not forwarded_for:
            request.state.forwarded_host = None
            request.state.forwarded_proto = None
            return await call_next(request)

        client = None
        for n, ip in enumerate(forwarded_for):
            if any(ip in network for network in self.proxies):
                continue
            client = str(ip)
            index = n
            break

        # If all the IP addresses are from trusted networks, take the
        # left-most.
        if not client:
            client = str(forwarded_for[-1])
            index = -1

        # Update the request's understanding of the client IP.  This uses an
        # undocumented interface; hopefully it will keep working.
        if request.client:
            request.scope["client"] = (client, request.client.port)
        else:
            request.scope["client"] = (client, None)

        # Ideally this should take the scheme corresponding to the entry in
        # X-Forwarded-For that was chosen, but some proxies (the Kubernetes
        # NGINX ingress, for example) only retain one element in
        # X-Forwarded-Proto.  In that case, use what we have.
        proto = list(reversed(self._get_forwarded_proto(request)))
        if proto:
            if index >= len(proto):
                index = -1
            request.scope["scheme"] = proto[index]

        # Rather than one entry per hop, NGINX seems to add only a single
        # X-Forwarded-Host header with the original hostname.
        request.state.forwarded_host = self._get_forwarded_host(request)

        return await call_next(request)

    def _get_forwarded_for(self, request: Request) -> list[_BaseAddress]:
        """Retrieve the ``X-Forwarded-For`` entries from the request.

        Parameters
        ----------
        request
            The incoming request.

        Returns
        -------
        list of ipaddress._BaseAddress
            The list of addresses found in the header.  If there are multiple
            ``X-Forwarded-For`` headers, we don't know which one is correct,
            so act as if there are no headers.
        """
        forwarded_for_str = request.headers.getlist("X-Forwarded-For")
        if not forwarded_for_str or len(forwarded_for_str) > 1:
            return []
        return [
            ip_address(addr)
            for addr in (a.strip() for a in forwarded_for_str[0].split(","))
            if addr
        ]

    def _get_forwarded_host(self, request: Request) -> str | None:
        """Retrieve the ``X-Forwarded-Host`` header.

        Parameters
        ----------
        request
            The incoming request.

        Returns
        -------
        str
            The value of the ``X-Forwarded-Host`` header, if present and if
            there is only one header.  If there are multiple
            ``X-Forwarded-Host`` headers, we don't know which one is correct,
            so act as if there are no headers.
        """
        forwarded_host = request.headers.getlist("X-Forwarded-Host")
        if not forwarded_host or len(forwarded_host) > 1:
            return None
        return forwarded_host[0].strip()

    def _get_forwarded_proto(self, request: Request) -> list[str]:
        """Retrieve the ``X-Forwarded-Proto`` entries from the request.

        Parameters
        ----------
        request
            The incoming request.

        Returns
        -------
        list of str
            The list of schemes found in the header.  If there are multiple
            ``X-Forwarded-Proto`` headers, we don't know which one is correct,
            so act as if there are no headers.
        """
        forwarded_proto_str = request.headers.getlist("X-Forwarded-Proto")
        if not forwarded_proto_str or len(forwarded_proto_str) > 1:
            return []
        return [p.strip() for p in forwarded_proto_str[0].split(",")]



"""Tests for the database utility functions."""

from __future__ import annotations

import os
from datetime import UTC, datetime, timedelta, timezone
from urllib.parse import unquote, urlparse

import pytest
import structlog
from sqlalchemy import Column, MetaData, String, Table
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.future import select
from sqlalchemy.orm import declarative_base

from safir.database import (
    _build_database_url,
    create_async_session,
    create_database_engine,
    create_sync_session,
    datetime_from_db,
    datetime_to_db,
    initialize_database,
)

TEST_DATABASE_URL = os.environ["TEST_DATABASE_URL"]
TEST_DATABASE_PASSWORD = os.getenv("TEST_DATABASE_PASSWORD")

Base = declarative_base()


class User(Base):
    """Tiny database table for testing."""

    __tablename__ = "user"

    username: str = Column(String(64), primary_key=True)


@pytest.mark.asyncio
async def test_database_init() -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(TEST_DATABASE_URL, TEST_DATABASE_PASSWORD)
    await initialize_database(engine, logger, schema=Base.metadata, reset=True)
    session = await create_async_session(engine, logger)
    async with session.begin():
        session.add(User(username="someuser"))
    await session.remove()

    # Reinitializing the database without reset should preserve the row.
    await initialize_database(engine, logger, schema=Base.metadata)
    session = await create_async_session(engine, logger)
    async with session.begin():
        result = await session.scalars(select(User.username))
        assert result.all() == ["someuser"]
    await session.remove()

    # Reinitializing the database with reset should delete the data.
    await initialize_database(engine, logger, schema=Base.metadata, reset=True)
    session = await create_async_session(engine, logger)
    async with session.begin():
        result = await session.scalars(select(User.username))
        assert result.all() == []
    await session.remove()
    await engine.dispose()


def test_build_database_url() -> None:
    url = _build_database_url(TEST_DATABASE_URL, None, is_async=False)
    assert url == TEST_DATABASE_URL

    url = _build_database_url(
        "postgresql://foo@127.0.0.1/foo", "password", is_async=False
    )
    assert url == "postgresql://foo:password@127.0.0.1/foo"

    url = _build_database_url(
        "postgresql://foo@127.0.0.1/foo", None, is_async=True
    )
    assert url == "postgresql+asyncpg://foo@127.0.0.1/foo"

    url = _build_database_url(
        "postgresql://foo@127.0.0.1:5432/foo", None, is_async=True
    )
    assert url == "postgresql+asyncpg://foo@127.0.0.1:5432/foo"

    url = _build_database_url(
        "postgresql://foo@127.0.0.1/foo", "otherpass", is_async=True
    )
    assert url == "postgresql+asyncpg://foo:otherpass@127.0.0.1/foo"

    url = _build_database_url(
        "postgresql://foo@127.0.0.1:5433/foo", "otherpass", is_async=True
    )
    assert url == "postgresql+asyncpg://foo:otherpass@127.0.0.1:5433/foo"

    # Test that the username and password are quoted properly.
    url = _build_database_url(
        "postgresql://foo%40e.com@127.0.0.1:4444/foo",
        "pass@word/with stuff",
        is_async=False,
    )
    assert url == (
        "postgresql://foo%40e.com:pass%40word%2Fwith%20stuff@127.0.0.1:4444"
        "/foo"
    )
    parsed_url = urlparse(url)
    assert parsed_url.username
    assert parsed_url.password

    # urlparse does not undo quoting in the components of netloc.
    assert unquote(parsed_url.username) == "foo@e.com"
    assert unquote(parsed_url.password) == "pass@word/with stuff"
    assert parsed_url.hostname == "127.0.0.1"
    assert parsed_url.port == 4444


@pytest.mark.asyncio
async def test_create_async_session() -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(TEST_DATABASE_URL, TEST_DATABASE_PASSWORD)
    await initialize_database(engine, logger, schema=Base.metadata, reset=True)

    session = await create_async_session(
        engine, logger, statement=select(User)
    )
    async with session.begin():
        session.add(User(username="foo"))
    await session.remove()

    # Use a query against a non-existent table as the liveness check and
    # ensure that fails.
    metadata = MetaData()
    bad_table = Table("bad", metadata, Column("name", String(64)))
    with pytest.raises(ProgrammingError):
        session = await create_async_session(
            engine, logger, statement=select(bad_table)
        )
    await engine.dispose()


@pytest.mark.asyncio
async def test_create_sync_session() -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(TEST_DATABASE_URL, TEST_DATABASE_PASSWORD)
    await initialize_database(engine, logger, schema=Base.metadata, reset=True)
    await engine.dispose()

    session = create_sync_session(
        TEST_DATABASE_URL,
        TEST_DATABASE_PASSWORD,
        logger,
        statement=select(User),
    )
    with session.begin():
        session.add(User(username="foo"))
    session.remove()

    # Use a query against a non-existent table as the liveness check and
    # ensure that fails.
    metadata = MetaData()
    bad_table = Table("bad", metadata, Column("name", String(64)))
    with pytest.raises(ProgrammingError):
        session = create_sync_session(
            TEST_DATABASE_URL,
            TEST_DATABASE_PASSWORD,
            logger,
            statement=select(bad_table),
        )


def test_datetime() -> None:
    tz_aware = datetime.now(tz=UTC)
    tz_naive = tz_aware.replace(tzinfo=None)

    assert datetime_to_db(tz_aware) == tz_naive
    assert datetime_from_db(tz_naive) == tz_aware
    assert datetime_from_db(tz_aware) == tz_aware

    assert datetime_to_db(None) is None
    assert datetime_from_db(None) is None

    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_to_db(tz_naive)

    tz_local = datetime.now(tz=timezone(timedelta(hours=1)))
    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_to_db(tz_local)
    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_from_db(tz_local)



"""Test the HTTP client FastAPI dependency."""

from __future__ import annotations

import pytest
import respx
from asgi_lifespan import LifespanManager
from fastapi import Depends, FastAPI
from httpx import AsyncClient

from safir.dependencies.http_client import http_client_dependency


@pytest.fixture
def non_mocked_hosts() -> list[str]:
    return ["example.com"]


@pytest.mark.asyncio
async def test_http_client(respx_mock: respx.Router) -> None:
    app = FastAPI()
    respx_mock.get("https://www.google.com").respond(200)

    @app.get("/")
    async def handler(
        http_client: AsyncClient = Depends(http_client_dependency),
    ) -> dict[str, str]:
        assert isinstance(http_client, AsyncClient)
        await http_client.get("https://www.google.com")
        return {}

    @app.on_event("shutdown")
    async def shutdown_event() -> None:
        await http_client_dependency.aclose()

    async with LifespanManager(app):
        async with AsyncClient(app=app, base_url="http://example.com") as c:
            r = await c.get("/")
    assert r.status_code == 200



"""Test IVOA middleware."""

from __future__ import annotations

import pytest
from fastapi import FastAPI
from httpx import AsyncClient

from safir.middleware.ivoa import CaseInsensitiveQueryMiddleware


def build_app() -> FastAPI:
    """Construct a test FastAPI app with the middleware registered."""
    app = FastAPI()
    app.add_middleware(CaseInsensitiveQueryMiddleware)
    return app


@pytest.mark.asyncio
async def test_case_insensitive() -> None:
    app = build_app()

    @app.get("/")
    async def handler(param: str) -> dict[str, str]:
        return {"param": param}

    async with AsyncClient(app=app, base_url="https://example.com") as client:
        r = await client.get("/", params={"param": "foo"})
        assert r.status_code == 200
        assert r.json() == {"param": "foo"}

        r = await client.get("/", params={"PARAM": "foo"})
        assert r.status_code == 200
        assert r.json() == {"param": "foo"}

        r = await client.get("/", params={"pARam": "foo"})
        assert r.status_code == 200
        assert r.json() == {"param": "foo"}

        r = await client.get("/", params={"paramX": "foo"})
        assert r.status_code == 422



"""Test the logger FastAPI dependency."""

from __future__ import annotations

import json
from unittest.mock import ANY

import pytest
from _pytest.logging import LogCaptureFixture
from fastapi import Depends, FastAPI
from httpx import AsyncClient
from structlog.stdlib import BoundLogger

from safir.dependencies.logger import logger_dependency
from safir.logging import configure_logging
from safir.middleware.x_forwarded import XForwardedMiddleware


@pytest.mark.asyncio
async def test_logger(caplog: LogCaptureFixture) -> None:
    configure_logging(name="myapp", profile="production", log_level="info")

    app = FastAPI()

    @app.get("/")
    async def handler(
        logger: BoundLogger = Depends(logger_dependency),
    ) -> dict[str, str]:
        logger.info("something", param="value")
        return {}

    caplog.clear()
    async with AsyncClient(app=app, base_url="http://example.com") as client:
        r = await client.get(
            "/", headers={"User-Agent": "some-user-agent/1.0"}
        )
    assert r.status_code == 200

    assert len(caplog.record_tuples) == 1
    assert json.loads(caplog.record_tuples[0][2]) == {
        "event": "something",
        "httpRequest": {
            "requestMethod": "GET",
            "requestUrl": "http://example.com/",
            "remoteIp": "127.0.0.1",
            "userAgent": "some-user-agent/1.0",
        },
        "logger": "myapp",
        "param": "value",
        "request_id": ANY,
        "severity": "info",
    }


@pytest.mark.asyncio
async def test_logger_xforwarded(caplog: LogCaptureFixture) -> None:
    configure_logging(name="myapp", profile="production", log_level="info")

    app = FastAPI()
    app.add_middleware(XForwardedMiddleware)

    @app.get("/")
    async def handler(
        logger: BoundLogger = Depends(logger_dependency),
    ) -> dict[str, str]:
        logger.info("something", param="value")
        return {}

    caplog.clear()
    async with AsyncClient(app=app, base_url="http://example.com") as client:
        r = await client.get(
            "/",
            headers={
                "Host": "foo.example.com",
                "User-Agent": "",
                "X-Forwarded-For": "10.10.10.10",
                "X-Forwarded-Proto": "https",
                "X-Forwarded-Host": "foo.example.com",
            },
        )
    assert r.status_code == 200

    assert len(caplog.record_tuples) == 1
    assert json.loads(caplog.record_tuples[0][2]) == {
        "event": "something",
        "httpRequest": {
            "requestMethod": "GET",
            "requestUrl": "https://foo.example.com/",
            "remoteIp": "10.10.10.10",
        },
        "logger": "myapp",
        "param": "value",
        "request_id": ANY,
        "severity": "info",
    }



"""Utility functions for database management."""

from __future__ import annotations

import asyncio
import time
from datetime import UTC, datetime
from typing import overload
from urllib.parse import quote, urlparse

from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    async_scoped_session,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.schema import CreateSchema
from sqlalchemy.sql.expression import Select
from sqlalchemy.sql.schema import MetaData
from structlog.stdlib import BoundLogger

__all__ = [
    "DatabaseInitializationError",
    "create_async_session",
    "create_database_engine",
    "create_sync_session",
    "datetime_from_db",
    "datetime_to_db",
    "initialize_database",
]


class DatabaseInitializationError(Exception):
    """Database initialization failed."""


def _build_database_url(
    url: str, password: str | None, *, is_async: bool
) -> str:
    """Build the authenticated URL for the database.

    Parameters
    ----------
    url
        Database connection URL, not including the password.
    password
        Database connection password.
    is_async
        Whether the resulting URL should be async or not.

    Returns
    -------
    url
        The URL including the password.

    Raises
    ------
    ValueError
        A password was provided but the connection URL has no username.
    """
    if is_async or password:
        parsed_url = urlparse(url)
        if is_async and parsed_url.scheme == "postgresql":
            parsed_url = parsed_url._replace(scheme="postgresql+asyncpg")
        if password:
            if not parsed_url.username:
                raise ValueError(f"No username in database URL {url}")
            password = quote(password, safe="")

            # The username portion of the parsed URL does not appear to decode
            # URL escaping of the username, so we should not quote it again or
            # we will get double-quoting.
            netloc = f"{parsed_url.username}:{password}@{parsed_url.hostname}"
            if parsed_url.port:
                netloc = f"{netloc}:{parsed_url.port}"
            parsed_url = parsed_url._replace(netloc=netloc)
        url = parsed_url.geturl()
    return url


@overload
def datetime_from_db(time: datetime) -> datetime:
    ...


@overload
def datetime_from_db(time: None) -> None:
    ...


def datetime_from_db(time: datetime | None) -> datetime | None:
    """Add the UTC time zone to a naive datetime from the database.

    Parameters
    ----------
    time
        The naive datetime from the database, or `None`.

    Returns
    -------
    datetime.datetime or None
        `None` if the input was none, otherwise a timezone-aware version of
        the same `~datetime.datetime` in the UTC timezone.
    """
    if not time:
        return None
    if time.tzinfo not in (None, UTC):
        raise ValueError(f"datetime {time} not in UTC")
    return time.replace(tzinfo=UTC)


@overload
def datetime_to_db(time: datetime) -> datetime:
    ...


@overload
def datetime_to_db(time: None) -> None:
    ...


def datetime_to_db(time: datetime | None) -> datetime | None:
    """Strip time zone for storing a datetime in the database.

    Parameters
    ----------
    time
        The timezone-aware `~datetime.datetime` in the UTC time zone, or
        `None`.

    Returns
    -------
    datetime.datetime or None
        `None` if the input was `None`, otherwise the same
        `~datetime.datetime` but timezone-naive and thus suitable for storing
        in a SQL database.
    """
    if not time:
        return None
    if time.tzinfo != UTC:
        raise ValueError(f"datetime {time} not in UTC")
    return time.replace(tzinfo=None)


def create_database_engine(
    url: str,
    password: str | None,
    *,
    isolation_level: str | None = None,
) -> AsyncEngine:
    """Create a new async database engine.

    Parameters
    ----------
    url
        Database connection URL, not including the password.
    password
        Database connection password.
    isolation_level
        If specified, sets a non-default isolation level for the database
        engine.

    Returns
    -------
    sqlalchemy.ext.asyncio.AsyncEngine
        Newly-created database engine.  When done with the engine, the caller
        must call ``await engine.dispose()``.

    Raises
    ------
    ValueError
        A password was provided but the connection URL has no username.
    """
    url = _build_database_url(url, password, is_async=True)
    if isolation_level:
        return create_async_engine(
            url, future=True, isolation_level=isolation_level
        )
    else:
        return create_async_engine(url, future=True)


async def create_async_session(
    engine: AsyncEngine,
    logger: BoundLogger | None = None,
    *,
    statement: Select | None = None,
) -> async_scoped_session:
    """Create a new async database session.

    Optionally checks that the database is available and retries in a loop for
    10s if it is not.  This should be used during application startup to wait
    for any network setup or database proxy sidecar.

    Parameters
    ----------
    engine
        Database engine to use for the session.
    logger
        Logger for reporting errors.  Used only if a statement is provided.
    statement
        If provided, statement to run to check database connectivity.  This
        will be modified with ``limit(1)`` before execution.  If not provided,
        database connectivity will not be checked.

    Returns
    -------
    sqlalchemy.ext.asyncio.async_scoped_session
        The database session proxy.  This is an asyncio scoped session that is
        scoped to the current task, which means that it will materialize new
        AsyncSession objects for each asyncio task (and thus each web
        request).  ``await session.remove()`` should be called when the caller
        is done with the session.
    """
    factory = async_sessionmaker(engine, expire_on_commit=False)
    session = async_scoped_session(factory, scopefunc=asyncio.current_task)

    # If no statement was provided, just return the async_scoped_session.
    if statement is None:
        return session

    # A statement was provided, so we want to check connectivity and retry for
    # up to ten seconds before returning the session.
    for _ in range(5):
        try:
            async with session.begin():
                await session.execute(statement.limit(1))
                return session
        except (ConnectionRefusedError, OperationalError, OSError):
            if logger:
                logger.info("database not ready, waiting two seconds")
            await session.remove()
            await asyncio.sleep(2)
            continue

    # If we got here, we failed five times.  Try one last time without
    # catching exceptions so that we raise the appropriate exception to our
    # caller.
    async with session.begin():
        await session.execute(statement.limit(1))
        return session


def create_sync_session(
    url: str,
    password: str | None,
    logger: BoundLogger | None = None,
    *,
    isolation_level: str | None = None,
    statement: Select | None = None,
) -> scoped_session:
    """Create a new sync database session.

    Used instead of `create_database_engine` and `create_async_session` for
    sync code, such as Dramatiq workers.  This combines engine creation with
    session creation.

    Parameters
    ----------
    url
        Database connection URL, not including the password.
    password
        Database connection password.
    logger
        Logger for reporting errors.  Used only if a statement is provided.
    isolation_level
        If specified, sets a non-default isolation level for the database
        engine.
    statement
        If provided, statement to run to check database connectivity.  This
        will be modified with ``limit(1)`` before execution.  If not provided,
        database connectivity will not be checked.

    Returns
    -------
    sqlalchemy.orm.scoping.scoped_session
        The database session proxy.  This manages a separate session per
        thread and therefore should be thread-safe.

    Raises
    ------
    ValueError
        A password was provided but the connection URL has no username.
    """
    url = _build_database_url(url, password, is_async=False)
    if isolation_level:
        engine = create_engine(url, isolation_level=isolation_level)
    else:
        engine = create_engine(url)
    factory = sessionmaker(bind=engine, future=True)
    session = scoped_session(factory)

    # If no statement was provided, just return the scoped_session.
    if statement is None:
        return session

    # A statement was provided, so we want to check connectivity and retry for
    # up to ten seconds before returning the session.
    for _ in range(5):
        try:
            with session.begin():
                session.execute(statement.limit(1))
                return session
        except (ConnectionRefusedError, OperationalError, OSError):
            if logger:
                logger.info("database not ready, waiting two seconds")
            session.remove()
            time.sleep(2)
            continue

    # If we got here, we failed five times.  Try one last time without
    # catching exceptions so that we raise the appropriate exception to our
    # caller.
    with session.begin():
        session.execute(statement.limit(1))
        return session


async def initialize_database(
    engine: AsyncEngine,
    logger: BoundLogger,
    *,
    schema: MetaData,
    reset: bool = False,
) -> None:
    """Create and initialize a new database.

    Parameters
    ----------
    engine
        Database engine to use.  Create with `create_database_engine`.
    logger
        Logger used to report problems
    schema
        Metadata for the database schema.  Generally this will be
        ``Base.metadata`` where ``Base`` is the declarative base used as the
        base class for all ORM table definitions.  The caller must ensure that
        all table definitions have been imported by Python before calling this
        function, or parts of the schema will be missing.
    reset
        If set to `True`, drop all tables and reprovision the database.
        Useful when running tests with an external database.

    Raises
    ------
    DatabaseInitializationError
        After five attempts, the database still could not be initialized.
        This is normally due to some connectivity issue to the database.
    """
    success = False
    error = None
    for _ in range(5):
        try:
            async with engine.begin() as conn:
                if schema.schema is not None:
                    await conn.execute(CreateSchema(schema.schema, True))
                if reset:
                    await conn.run_sync(schema.drop_all)
                await conn.run_sync(schema.create_all)
            success = True
        except (ConnectionRefusedError, OperationalError, OSError) as e:
            logger.info("database not ready, waiting two seconds")
            error = str(e)
            await asyncio.sleep(2)
            continue
        if success:
            logger.info("initialized database schema")
            break
    if not success:
        msg = "database schema initialization failed (database not reachable?)"
        logger.error(msg)
        await engine.dispose()
        raise DatabaseInitializationError(error)
